# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated hypothesis-based clustering implementation.

Based on the paper:

Three Approaches for Personalization with Applications to Federated Learning
    Yishay Mansour, Mehryar Mohri, Jae Ro, Ananda Theertha Suresh
    https://arxiv.org/abs/2002.10619
"""

from typing import Iterator, List, Mapping, NamedTuple, Optional

from fedjax.legacy import core
import jax.numpy as jnp
import jax.random as jrandom
import numpy as np


class HypClusterState(NamedTuple):
  """State of server for HypCluster passed between rounds.

  Attributes:
    cluster_params: A list of pytrees representing the server model parameters
      per cluster.
    cluster_server_opt_states: A list of pytrees representing the server
      optimizer state per cluster.
  """
  cluster_params: List[core.Params]
  cluster_server_opt_states: List[core.OptState]


class HypClusterHParams(NamedTuple):
  """Hyperparameters for federated hypothesis-based clustering.

  Attributes:
    train_data_hparams: Hyperparameters for training client data preparation.
    num_clusters: Number of clusters.
    init_cluster_client_ids: Optional list of client ids used to initialize
      cluster centers via k-means++ variant. If None, initialization is random.
  """
  train_data_hparams: core.ClientDataHParams
  num_clusters: int = 1
  init_cluster_client_ids: Optional[List[str]] = None


def _get_cluster_by_client_loss(
    federated_data: core.FederatedData,
    client_ids: List[str],
    model: core.Model,
    cluster_params: List[core.Params],
    data_hparams: Optional[core.ClientDataHParams] = None) -> np.ndarray:
  """Returns the loss per cluster and client pair.

  Args:
    federated_data: Federated data separated per client to be evaluated.
    client_ids: Ids of clients to be evaluated.
    model: Model implementation.
    cluster_params: List of pytree model parameters per cluster.
    data_hparams: Hyperparameters for dataset preparation.

  Returns:
    Numpy array of losses of shape [num_clusters, num_clients].
  """
  cluster_by_client_loss = []
  for params in cluster_params:
    client_metrics = core.evaluate_multiple_clients(
        federated_data=federated_data,
        client_ids=client_ids,
        model=model,
        params=params,
        client_data_hparams=data_hparams)
    client_loss = [m['loss'] if m else np.inf for m in client_metrics]
    cluster_by_client_loss.append(client_loss)
  # [num_clusters, num_clients].
  cluster_by_client_loss = np.array(cluster_by_client_loss)
  return cluster_by_client_loss


def maximization(
    federated_data: core.FederatedData,
    client_ids: List[str],
    model: core.Model,
    cluster_params: List[core.Params],
    data_hparams: Optional[core.ClientDataHParams] = None) -> List[List[str]]:
  """Finds the cluster that produces the lowest loss for each client.

  Args:
    federated_data: Federated data separated per client to be evaluated.
    client_ids: Ids of clients to be evaluated.
    model: Model implementation.
    cluster_params: List of pytree model parameters per cluster.
    data_hparams: Hyperparameters for dataset preparation.

  Returns:
    List of lists of client ids per cluster.
  """
  cluster_by_client_loss = _get_cluster_by_client_loss(federated_data,
                                                       client_ids, model,
                                                       cluster_params,
                                                       data_hparams)
  # [num_clients] index of best cluster per client.
  cluster_ids = np.argmin(cluster_by_client_loss, axis=0)

  cluster_client_ids = [[] for _ in cluster_params]
  for cluster_id, client_id in zip(cluster_ids, client_ids):
    cluster_client_ids[cluster_id].append(client_id)
  return cluster_client_ids


# TODO(b/162864512): Break up functions to jit compile.
class HypCluster(core.FederatedAlgorithm):
  """Federated hypothesis-based clustering algorithm."""

  def __init__(self, federated_data: core.FederatedData, model: core.Model,
               client_optimizer: core.Optimizer,
               server_optimizer: core.Optimizer, hparams: HypClusterHParams,
               rng_seq: core.PRNGSequence):
    """Initializes HypCluster algorithm.

    Args:
      federated_data: Federated data separated per client.
      model: Model implementation.
      client_optimizer: Client optimizer.
      server_optimizer: Server optimizer.
      hparams: Hyperparameters for federated hypothesis-based clustering.
      rng_seq: Iterator of JAX random keys.
    """
    self._federated_data = federated_data
    self._model = model
    self._client_optimizer = client_optimizer
    self._server_optimizer = server_optimizer
    self._hparams = hparams
    self._rng_seq = rng_seq
    self._client_trainer = core.DefaultClientTrainer(model, client_optimizer)

  def _kmeans_init(self) -> List[core.Params]:
    """Initializes cluster parameters for HypCluster using k-means++ variant.

    Given a set of input clients, we train parameters for each client. The
    initial cluster parameters are chosen out of this set of client parameters.
    At a high level, the cluster parameters are selected by choosing the client
    parameters that are the "furthest away" from each other (greatest difference
    in loss).

    Returns:
      Cluster center params.
    """
    init_params = self._model.init_params(next(self._rng_seq))
    client_ids = self._hparams.init_cluster_client_ids
    client_states = core.train_multiple_clients(
        federated_data=self.federated_data,
        client_ids=client_ids,
        client_trainer=self._client_trainer,
        init_client_trainer_state=self._client_trainer.init_state(
            params=init_params),
        rng_seq=self._rng_seq,
        client_data_hparams=self._hparams.train_data_hparams)
    params_by_client = [cs.params for cs in client_states]

    center_client = jrandom.choice(next(self._rng_seq), len(params_by_client))
    cluster_centers = [params_by_client[center_client]]

    for _ in range(1, self._hparams.num_clusters):
      # [num_centers, num_clients].
      cluster_by_client_loss = _get_cluster_by_client_loss(
          self.federated_data, client_ids, self.model, cluster_centers,
          self._hparams.train_data_hparams)
      # [num_clients].
      best_loss_per_client = np.min(cluster_by_client_loss, axis=0)
      worst_client = np.argmax(best_loss_per_client)
      cluster_centers.append(params_by_client[worst_client])

    return cluster_centers

  def _expectation(self, cluster_params: List[core.Params],
                   cluster_client_ids: List[List[str]]) -> List[core.Params]:
    """Trains each cluster's parameters on the clients in that cluster.

    Args:
      cluster_params: List of pytree model parameters per cluster.
      cluster_client_ids: List of lists of client ids per cluster.

    Returns:
      List of model parameter deltas per cluster.
    """
    cluster_delta_params = []
    for client_ids, params in zip(cluster_client_ids, cluster_params):
      if not client_ids:
        # Empty clusters shouldn't contribute to training since they have no
        # training data.
        cluster_delta_params.append(core.tree_util.tree_zeros_like(params))
        continue

      # Train cluster clients.
      client_states = core.train_multiple_clients(
          federated_data=self.federated_data,
          client_ids=client_ids,
          client_trainer=self._client_trainer,
          init_client_trainer_state=self._client_trainer.init_state(params),
          rng_seq=self._rng_seq,
          client_data_hparams=self._hparams.train_data_hparams)

      # Weighted average of param delta across clients.
      def select_delta_params_and_weight(client_state, p=params):
        delta_params = core.tree_multimap(lambda a, b: a - b, p,
                                          client_state.params)
        return delta_params, client_state.num_examples

      delta_params_and_weight = map(select_delta_params_and_weight,
                                    client_states)
      delta_params = core.tree_mean(delta_params_and_weight)
      cluster_delta_params.append(delta_params)
    return cluster_delta_params

  @property
  def federated_data(self) -> core.FederatedData:
    return self._federated_data

  @property
  def model(self) -> core.Model:
    return self._model

  def init_state(self) -> HypClusterState:
    if self._hparams.init_cluster_client_ids:
      cluster_params = self._kmeans_init()
    else:
      cluster_params = [
          self._model.init_params(next(self._rng_seq))
          for _ in range(self._hparams.num_clusters)
      ]

    cluster_server_opt_states = [
        self._server_optimizer.init_fn(p) for p in cluster_params
    ]
    return HypClusterState(
        cluster_params=cluster_params,
        cluster_server_opt_states=cluster_server_opt_states)

  def run_round(self, state: HypClusterState,
                client_ids: List[str]) -> HypClusterState:
    """Runs federated hypothesis-based clustering."""
    cluster_client_ids = maximization(self.federated_data, client_ids,
                                      self.model, state.cluster_params,
                                      self._hparams.train_data_hparams)
    cluster_delta_params = self._expectation(state.cluster_params,
                                             cluster_client_ids)

    cluster_params = state.cluster_params
    cluster_server_opt_states = state.cluster_server_opt_states
    for cluster_id, client_ids in enumerate(cluster_client_ids):
      if not client_ids:
        # Empty clusters shouldn't contribute to training since they have no
        # training data.
        continue
      updates, opt_state = self._server_optimizer.update_fn(
          cluster_delta_params[cluster_id],
          state.cluster_server_opt_states[cluster_id])
      cluster_server_opt_states[cluster_id] = opt_state
      cluster_params[cluster_id] = self._server_optimizer.apply_updates(
          state.cluster_params[cluster_id], updates)

    return HypClusterState(
        cluster_params=cluster_params,
        cluster_server_opt_states=cluster_server_opt_states)


def evaluate_multiple_clients_with_clusters(
    federated_data: core.FederatedData,
    client_ids: List[str],
    model: core.Model,
    cluster_params: List[core.Params],
    data_hparams: Optional[core.ClientDataHParams] = None
) -> Iterator[Mapping[str, jnp.ndarray]]:
  """Evaluates cluster parameters over input clients' datasets.

  Args:
    federated_data: Federated data separated per client to be evaluated.
    client_ids: Ids of clients to be evaluated.
    model: Model implementation.
    cluster_params: List of pytree model parameters per cluster.
    data_hparams: Hyperparameters for client dataset preparation.

  Yields:
    Ordered mapping of metric names to values per client.
  """
  cluster_client_ids = maximization(federated_data, client_ids, model,
                                    cluster_params, data_hparams)
  for client_ids_, params in zip(cluster_client_ids, cluster_params):
    client_metrics = core.evaluate_multiple_clients(
        federated_data=federated_data,
        client_ids=client_ids_,
        model=model,
        params=params,
        client_data_hparams=data_hparams)
    yield from client_metrics



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Algorithms for model update compression."""

from typing import Iterable, Tuple

from fedjax.aggregators import aggregator
from fedjax.core import dataclasses
from fedjax.core import tree_util
from fedjax.core.federated_data import ClientId
from fedjax.core.typing import PRNGKey, Params
import haiku as hk
import jax
import jax.numpy as jnp


@dataclasses.dataclass
class CompressionState:
  """State of compression aggregator passed between rounds.

  Attributes:
    num_bits: number of bits transmitted per client.
    rng: random key used for compression.
  """
  num_bits: float
  rng: PRNGKey


def binary_stochastic_quantize(v, rng):
  """Binary stochastic algorithm in https://arxiv.org/pdf/1611.00429.pdf.

  Args:
    v: vector to be quantized.
    rng: jax random key.

  Returns:
    Quantized vector.
  """
  v_min = jnp.amin(v)
  v_max = jnp.amax(v)
  v = jnp.nan_to_num((v - v_min) / (v_max - v_min))
  rand = jax.random.uniform(key=rng, shape=v.shape)
  return jnp.where(rand > v, v_min, v_max)


def uniform_stochastic_quantize(v, num_levels, rng):
  """Uniform stochastic algorithm in https://arxiv.org/pdf/1611.00429.pdf.

  Args:
    v: vector to be quantized.
    num_levels: Number of levels of quantization.
    rng: jax random key.

  Returns:
    Quantized vector.
  """
  # Rescale the vector to be between zero to one.
  v_min = jnp.amin(v)
  v_max = jnp.amax(v)
  v = jnp.nan_to_num((v - v_min) / (v_max - v_min))
  # Compute the upper and lower boundary of each value.
  v_ceil = jnp.ceil(v * (num_levels - 1)) / (num_levels - 1)
  v_floor = jnp.floor(v * (num_levels - 1)) / (num_levels - 1)
  # uniformly quantize between v_ceil and v_floor.
  rand = jax.random.uniform(key=rng, shape=v.shape)
  threshold = jnp.nan_to_num((v - v_floor) / (v_ceil - v_floor))
  quantized = jnp.where(rand > threshold, v_floor, v_ceil)
  # Rescale the values and return it.
  return v_min + quantized * (v_max - v_min)


@jax.jit
def uniform_stochastic_quantize_pytree(param, num_levels, rng):
  """Applies uniform_stochastic_quantize for all leafs of the pytree.

  Args:
    param: pytree to be quantized.
    num_levels: Number of levels of quantization.
    rng: jax random key.

  Returns:
    Quantized pytree.
  """
  leaves, tree_def = jax.tree_util.tree_flatten(param)
  rngs = jax.random.split(rng, len(leaves))
  new_leaves = []
  for i, r in zip(leaves, rngs):
    new_leaves.append(uniform_stochastic_quantize(i, num_levels, r))
  return jax.tree_util.tree_unflatten(tree_def, new_leaves)


def num_leaves(pytree):
  return len(jax.tree_util.tree_leaves(pytree))


def uniform_stochastic_quantizer(num_levels: int,
                                 rng: PRNGKey) -> aggregator.Aggregator:
  """Returns (weighted) mean of input uniformly quantized trees using the
  uniform stochastic algorithm in https://arxiv.org/pdf/1611.00429.pdf.

  Args:
    num_levels: number of levels of quantization.
    rng: PRNGKey used for compression.

  Returns:
    Compression aggreagtor.
  """

  def init():
    return CompressionState(0.0, rng)

  def apply(
      clients_params_and_weights: Iterable[Tuple[ClientId, Params, float]],
      aggregator_state: CompressionState) -> Tuple[Params, CompressionState]:

    def quantize_params_and_weight(clients_params_and_weight_rng):
      (_, param, weight), rng = clients_params_and_weight_rng

      return uniform_stochastic_quantize_pytree(param, num_levels, rng), weight

    rng, use_rng = jax.random.split(aggregator_state.rng)
    # TODO(theertha): remove the usage of hk.PRNGSequence.
    rng_seq = hk.PRNGSequence(use_rng)
    clients_params_and_weight_rng = zip(clients_params_and_weights, rng_seq)
    quantized_p_and_w = map(quantize_params_and_weight,
                            clients_params_and_weight_rng)
    aggregated_params = tree_util.tree_mean(quantized_p_and_w)
    total_num_params = tree_util.tree_size(aggregated_params)
    total_num_floats = num_leaves(aggregated_params)
    # 32 bits for every float used and one bit for every parameter.
    new_bits = total_num_params + 64 * total_num_floats
    new_state = CompressionState(aggregator_state.num_bits + new_bits, rng)
    return aggregated_params, new_state

  return aggregator.Aggregator(init, apply)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.optimizers."""

from absl.testing import absltest

from fedjax.core import optimizers

import flax
import haiku as hk
import jax
import jax.numpy as jnp
import numpy.testing as npt


class OptimizersTest(absltest.TestCase):

  def test_ignore_grads_haiku(self):
    params = hk.data_structures.to_immutable_dict({
        'linear_1': {
            'w': jnp.array([1., 1., 1.])
        },
        'linear_2': {
            'w': jnp.array([2., 2., 2.]),
            'b': jnp.array([3., 3., 3.])
        }
    })
    grads = jax.tree_util.tree_map(lambda _: 0.5, params)
    ignore_optimizer = optimizers.ignore_grads_haiku(
        optimizer=optimizers.sgd(learning_rate=1.0),
        non_trainable_names=[('linear_1', 'w'), ('linear_2', 'b')])

    opt_state = ignore_optimizer.init(params)
    opt_state, updated_params = ignore_optimizer.apply(grads, opt_state, params)

    jax.tree_util.tree_multimap(
        npt.assert_array_equal, updated_params,
        hk.data_structures.to_immutable_dict({
            'linear_1': {
                'w': jnp.array([1., 1., 1.])
            },
            'linear_2': {
                'w': jnp.array([1.5, 1.5, 1.5]),
                'b': jnp.array([3., 3., 3.])
            }
        }))

  def test_create_optimizer_from_flax(self):

    def create_optimizer_from_flax(opt_def, **hyper_param_overrides):
      hyper_params = opt_def.update_hyper_params(**hyper_param_overrides)

      def init(params):
        params = hk.data_structures.to_mutable_dict(params)
        return opt_def.init_state(params)

      @jax.jit
      def apply(grads, opt_state, params):
        # flax.optim doesn't play well with hk.FlatMapping.
        grads = hk.data_structures.to_mutable_dict(grads)
        params = hk.data_structures.to_mutable_dict(params)
        params, opt_state = opt_def.apply_gradient(hyper_params, params,
                                                   opt_state, grads)
        params = hk.data_structures.to_immutable_dict(params)
        return opt_state, params

      return optimizers.Optimizer(init, apply)

    params = hk.data_structures.to_immutable_dict(
        {'linear': {
            'w': jnp.array([1., 1., 1.])
        }})
    grads = jax.tree_util.tree_map(lambda _: 0.5, params)
    opt_def = flax.optim.Adam(learning_rate=0.1)
    optimizer = create_optimizer_from_flax(opt_def)

    opt_state = optimizer.init(params)
    opt_state, updated_params = optimizer.apply(grads, opt_state, params)

    # numpy.testing doesn't have an assert_array_not_equal.
    with npt.assert_raises(AssertionError):
      npt.assert_array_equal(updated_params['linear']['w'],
                             params['linear']['w'])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.client_samplers."""

from absl.testing import absltest

from fedjax.core import client_datasets
from fedjax.core import client_samplers

import jax
import numpy as np
import numpy.testing as npt


class ClientSamplersTest(absltest.TestCase):

  def assert_clients_equal(self, actual_clients, expected_clients):
    with self.subTest('client_ids'):
      self.assertEqual([cid for cid, _, _ in actual_clients],
                       [cid for cid, _, _ in expected_clients])
    with self.subTest('client_datasets'):
      self.assertEqual([len(cds) for _, cds, _ in actual_clients],
                       [len(cds) for _, cds, _ in expected_clients])
    with self.subTest('client_rngs'):
      npt.assert_array_equal([crng for _, _, crng in actual_clients],
                             [crng for _, _, crng in expected_clients])

  def test_uniform_shuffled_client_sampler(self):

    def shuffled_clients():
      i = 0
      while True:
        yield i, client_datasets.ClientDataset({'x': np.arange(i)})
        i += 1

    with self.subTest('sample'):
      client_sampler1 = client_samplers.UniformShuffledClientSampler(
          shuffled_clients(), num_clients=2)
      for _ in range(4):
        clients1 = client_sampler1.sample()
      client_sampler2 = client_samplers.UniformShuffledClientSampler(
          shuffled_clients(), num_clients=2, start_round_num=3)
      self.assert_clients_equal(client_sampler2.sample(), clients1)

    with self.subTest('set_round_num'):
      with self.assertRaisesRegex(NotImplementedError,
                                  '.*Use UniformGetClientSampler.*'):
        client_sampler1.set_round_num(1)

  def test_uniform_get_client_sampler(self):

    class MockFederatedData:

      def client_ids(self):
        yield from range(100)

      def get_clients(self, client_ids):
        for cid in client_ids:
          yield cid, client_datasets.ClientDataset({'x': np.arange(int(cid))})

    num_clients = 2
    round_num = 3
    client_sampler = client_samplers.UniformGetClientSampler(
        MockFederatedData(), num_clients, seed=0, start_round_num=round_num)
    with self.subTest('sample'):
      client_rngs = jax.random.split(jax.random.PRNGKey(round_num), num_clients)
      expect = [(78, client_datasets.ClientDataset({'x': np.arange(78)}),
                 client_rngs[0]),
                (56, client_datasets.ClientDataset({'x': np.arange(56)}),
                 client_rngs[1])]
      self.assert_clients_equal(client_sampler.sample(), expect)

    with self.subTest('set_round_num'):
      self.assertNotEqual(client_sampler._round_num, round_num)
      client_sampler.set_round_num(round_num)
      self.assert_clients_equal(client_sampler.sample(), expect)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mime Lite implementation.

Based on the paper:

Mime: Mimicking Centralized Stochastic Algorithms in Federated Learning
    Sai Praneeth Karimireddy, Martin Jaggi, Satyen Kale, Mehryar Mohri,
    Sashank J. Reddi, Sebastian U. Stich, Ananda Theertha Suresh
    https://arxiv.org/abs/2008.03606
"""

import functools
from typing import Iterable, List, NamedTuple

from fedjax.legacy import core
import jax
import jax.numpy as jnp


class MimeLiteState(NamedTuple):
  """State of server for Mime/MimeLite passed between rounds.

  Attributes:
    params: A pytree representing the server model parameters.
    opt_state: A pytree representing the base optimizer state.
  """
  params: core.Params
  opt_state: core.OptState


class MimeLiteHParams(NamedTuple):
  """Hyperparameters for Mime Lite algorithm.

  Attributes:
    train_data_hparams: Hyperparameters for training client data preparation.
    combined_data_hparams: Hyperparameters for training data preparation.
    server_learning_rate: Server learning rate scales the update from clients
      before applying to server. Defaults to 1.
  """
  train_data_hparams: core.ClientDataHParams
  combined_data_hparams: core.ClientDataHParams
  server_learning_rate: float = 1.0


class MimeLiteClientTrainer(core.DefaultClientTrainer):
  """Single client trainer that doesn't update optimizer state."""

  @functools.partial(jax.jit, static_argnums=0)
  def one_step(self, client_trainer_state: core.DefaultClientTrainerState,
               batch: core.Batch,
               rng: core.PRNGKey) -> core.DefaultClientTrainerState:
    backward_pass_output = self._model.backward_pass(
        client_trainer_state.params, batch, rng)
    params_updates, _ = self._optimizer.update_fn(
        backward_pass_output.grads, client_trainer_state.opt_state)
    params = self._optimizer.apply_updates(client_trainer_state.params,
                                           params_updates)
    return core.DefaultClientTrainerState(
        params=params,
        opt_state=client_trainer_state.opt_state,
        num_examples=client_trainer_state.num_examples +
        backward_pass_output.num_examples)


def compute_gradient(stream: Iterable[core.Batch], params: core.Params,
                     model: core.Model,
                     rng_seq: core.PRNGSequence) -> core.Updates:
  """Computes the gradient over the full stream at params."""
  num_examples = 0.
  grads_sum = jax.tree_map(jnp.zeros_like, params)
  for batch, rng in zip(stream, rng_seq):
    backward_pass_output = model.backward_pass(params, batch, rng)
    grads = backward_pass_output.grads
    batch_num_examples = backward_pass_output.num_examples
    grads_sum = jax.tree_multimap(
        lambda p, q, bs=batch_num_examples: p * bs + q, grads, grads_sum)
    num_examples += batch_num_examples
  return jax.tree_map(lambda p: p / num_examples, grads_sum)


class MimeLite(core.FederatedAlgorithm):
  """Mime Lite algorithm."""

  def __init__(self,
               federated_data: core.FederatedData,
               model: core.Model,
               base_optimizer: core.Optimizer,
               hparams: MimeLiteHParams,
               rng_seq: core.PRNGSequence):
    """Initializes MimeLite algorithm.

    Args:
      federated_data: Federated data separated per client.
      model: Model implementation.
      base_optimizer: Base centralized optimizer which we will try mimic.
      hparams: Hyperparameters for federated averaging.
      rng_seq: Iterator of JAX random keys.
    """
    self._federated_data = federated_data
    self._model = model
    self._base_optimizer = base_optimizer
    self._hparams = hparams
    self._rng_seq = rng_seq
    self._client_trainer = MimeLiteClientTrainer(model, base_optimizer)

  @property
  def federated_data(self) -> core.FederatedData:
    return self._federated_data

  @property
  def model(self) -> core.Model:
    return self._model

  def init_state(self) -> MimeLiteState:
    params = self._model.init_params(next(self._rng_seq))
    opt_state = self._base_optimizer.init_fn(params)
    return MimeLiteState(params=params, opt_state=opt_state)

  def run_round(self, state: MimeLiteState,
                client_ids: List[str]) -> MimeLiteState:
    """Runs one round of MimeLite."""
    # Train model per client.
    client_states = core.train_multiple_clients(
        federated_data=self.federated_data,
        client_ids=client_ids,
        client_trainer=self._client_trainer,
        init_client_trainer_state=self._client_trainer.init_state(
            state.params, state.opt_state),
        rng_seq=self._rng_seq,
        client_data_hparams=self._hparams.train_data_hparams)

    # Weighted average of param delta across clients.
    def select_delta_params_and_weight(client_state):
      delta_params = core.tree_multimap(lambda a, b: a - b, state.params,
                                        client_state.params)
      return delta_params, client_state.num_examples

    delta_params_and_weight = map(select_delta_params_and_weight, client_states)
    delta_params = core.tree_mean(delta_params_and_weight)

    # Compute full-batch gradient at server params on train data.
    combined_dataset = core.create_tf_dataset_for_clients(
        self.federated_data, client_ids)
    combined_dataset = core.preprocess_tf_dataset(
        combined_dataset, self._hparams.combined_data_hparams)
    server_grads = compute_gradient(
        stream=combined_dataset.as_numpy_iterator(),
        params=state.params,
        model=self._model,
        rng_seq=self._rng_seq,
    )

    # Server params uses weighted average of client updates, scaled by the
    # server_learning_rate.
    params = core.tree_multimap(
        lambda p, q: p - self._hparams.server_learning_rate * q, state.params,
        delta_params)
    # Update server opt_state using base_optimizer and server gradient.
    _, opt_state = self._base_optimizer.update_fn(server_grads, state.opt_state)
    return MimeLiteState(params=params, opt_state=opt_state)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated Shakespeare."""

import functools
from typing import Optional, Tuple

from fedjax.core import client_datasets
from fedjax.core import federated_data
from fedjax.core import sqlite_federated_data
from fedjax.datasets import downloads
import numpy as np

SPLITS = ('train', 'test')


def cite():
  return """@inproceedings{mcmahan2017communication,
  title={Communication-efficient learning of deep networks from
decentralized data},
  author={McMahan, Brendan and Moore, Eider and Ramage, Daniel and
Hampson, Seth and y Arcas, Blaise Aguera},
  booktitle={Artificial Intelligence and Statistics},
  pages={1273--1282},
  year={2017},
  organization={PMLR}
}"""


def load_split(split: str,
               mode: str = 'sqlite',
               cache_dir: Optional[str] = None) -> federated_data.FederatedData:
  """Loads a shakespeare split.

  Features:
    snippets: [N] bytes array of snippet text.

  Args:
    split: Name of the split. One of SPLITS.
    mode: 'sqlite'.
    cache_dir: Directory to cache files in 'sqlite' mode.

  Returns:
    FederatedData.
  """
  if split not in SPLITS:
    raise ValueError(f'Invalid split={split!r}')
  if cache_dir is not None and mode != 'sqlite':
    raise ValueError('Caching locally is only supported in "sqlite" mode')
  if mode == 'sqlite':
    path = downloads.maybe_download(
        f'https://storage.googleapis.com/gresearch/fedjax/shakespeare/shakespeare_{split}.sqlite',
        cache_dir)
    return sqlite_federated_data.SQLiteFederatedData.new(path)
  else:
    raise ValueError(f'Unsupported mode={mode!r}')


def load_data(
    sequence_length: int = 80,
    mode: str = 'sqlite',
    cache_dir: Optional[str] = None
) -> Tuple[federated_data.FederatedData, federated_data.FederatedData]:
  """Loads preprocessed shakespeare splits.

  Features (M below is possibly different from N in load_split):
    x: [M, sequence_length] int32 input labels, in the range of [0,
      shakespeare.VOCAB_SIZE)
    y: [M, sequence_length] int32 output labels, in the range of [0,
      shakespeare.VOCAB_SIZE)

  Args:
    sequence_length: The fixed sequence length after preprocessing.
    mode: 'sqlite'.
    cache_dir: Directory to cache files in 'sqlite' mode.

  Returns:
    A (train, held_out, test) tuple of federated data.
  """
  train = load_split('train', mode, cache_dir)
  test = load_split('test', mode, cache_dir)
  preprocess = functools.partial(
      preprocess_client, sequence_length=sequence_length)
  return (train.preprocess_client(preprocess),
          test.preprocess_client(preprocess))


def _build_look_up_table(vocab: bytes,
                         num_reserved: int) -> Tuple[np.ndarray, int]:
  """Builds a look-up table from a byte to its integer label.

  Args:
    vocab: bytes object listing the byte values to include in the vocabulary.
      The byte vocab[i] is assigned label `num_reserved + i`. If the same byte
      occurs multiple times, the index of the last occurrence is used.
    num_reserved: Number of labels to reserve in the beginning of the integer
      label domain. Bytes in `vocab` will not be mapped to [0, num_reserved).

  Returns:
    (table, vocab_size) tuple. `table` is simply a [256] ndarray containing the
    integer label for each byte. Bytes not in `vocab` are mapped to `vocab_size
    - 1`.
  """
  oov = num_reserved + len(vocab)
  vocab_size = oov + 1
  table = np.full([256], oov, dtype=np.int32)
  for i, c in enumerate(vocab):
    table[c] = num_reserved + i
  return table, vocab_size


# Vocabulary re-used from the Federated Learning for Text Generation tutorial.
# https://www.tensorflow.org/federated/tutorials/federated_learning_for_text_generation
TABLE, VOCAB_SIZE = _build_look_up_table(
    b'dhlptx@DHLPTX $(,048cgkoswCGKOSW[_#\'/37;?bfjnrvzBFJNRVZ"&*.26:\naeimquyAEIMQUY]!%)-159\r',
    num_reserved=3)
OOV = VOCAB_SIZE - 1
# Reserved labels.
PAD = 0
BOS = 1
EOS = 2


def preprocess_client(client_id: federated_data.ClientId,
                      examples: client_datasets.Examples,
                      sequence_length: int) -> client_datasets.Examples:
  """Turns snippets into sequences of integer labels.

  Features (M below is possibly different from N in load_split):
    x: [M, sequence_length] int32 input labels, in the range of [0,
      shakespeare.VOCAB_SIZE)
    y: [M, sequence_length] int32 output labels, in the range of [0,
      shakespeare.VOCAB_SIZE)

  All snippets in a client dataset are first joined into a single sequence (with
  BOS/EOS added), and then split into pairs of `sequence_length` chunks for
  language model training. For example, with sequence_length=3,
  `[b'ABCD', b'E']` becomes
  ```
  Input sequences:  [[BOS, A, B], [C, D, EOS],   [BOS, E, PAD]]
  Output seqeunces: [[A, B, C],   [D, EOS, BOS], [E, EOS, PAD]]
  ```

  Note: This is not equivalent to
  https://www.tensorflow.org/federated/tutorials/federated_learning_for_text_generation#load_and_preprocess_the_federated_shakespeare_data
  (The processing logic there loses ~1/sequence_length portion of the tokens).

  Args:
    client_id: Not used.
    examples: Unprocessed examples (e.g. from `load_split()`).
    sequence_length: The fixed sequence length after preprocessing.

  Returns:
    Processed examples.
  """
  del client_id
  snippets = examples['snippets']
  # Join all snippets into a single label sequence.
  joined_length = sum(len(i) + 2 for i in snippets)
  joined = np.zeros([joined_length], dtype=np.int32)
  offset = 0
  for i in snippets:
    joined[offset] = BOS
    joined[offset + 1:offset + 1 + len(i)] = TABLE[list(i)]
    joined[offset + 1 + len(i)] = EOS
    offset += len(i) + 2
  # Split into input/output sequences of size `sequence_length`.
  padded_length = ((joined_length - 1 + sequence_length - 1) //
                   sequence_length * sequence_length)
  input_labels = np.full([padded_length], PAD, dtype=np.int32)
  input_labels[:joined_length - 1] = joined[:-1]
  output_labels = np.full([padded_length], PAD, dtype=np.int32)
  output_labels[:joined_length - 1] = joined[1:]
  return {
      'x': input_labels.reshape([-1, sequence_length]),
      'y': output_labels.reshape([-1, sequence_length])
  }



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Registry of standard tasks.

Each task is represented as a (train federated data, test federated data, model)
tuple.
"""

from typing import Optional, Tuple

from fedjax import datasets
from fedjax import models
from fedjax.core import federated_data
from fedjax.core import models as core_models

ALL_TASKS = ('EMNIST_CONV', 'EMNIST_LOGISTIC', 'EMNIST_DENSE',
             'SHAKESPEARE_CHARACTER', 'STACKOVERFLOW_WORD')


def get_task(
    name: str,
    mode: str = 'sqlite',
    cache_dir: Optional[str] = None
) -> Tuple[federated_data.FederatedData, federated_data.FederatedData,
           core_models.Model]:
  """Gets a standard task.

  Args:
    name: Name of the task to get. Must be one of fedjax.training.ALL_TASKS.
    mode: 'sqlite'.
    cache_dir: Directory to cache files in 'sqlite' mode.

  Returns:
    (train federated data, test federated data, model) tuple.
  """
  if name == 'EMNIST_CONV':
    train, test = datasets.emnist.load_data(
        only_digits=False, mode=mode, cache_dir=cache_dir)
    model = models.emnist.create_conv_model(only_digits=False)
  elif name == 'EMNIST_LOGISTIC':
    train, test = datasets.emnist.load_data(
        only_digits=False, mode=mode, cache_dir=cache_dir)
    model = models.emnist.create_logistic_model(only_digits=False)
  elif name == 'EMNIST_DENSE':
    train, test = datasets.emnist.load_data(
        only_digits=False, mode=mode, cache_dir=cache_dir)
    model = models.emnist.create_dense_model(only_digits=False)
  elif name == 'SHAKESPEARE_CHARACTER':
    train, test = datasets.shakespeare.load_data(mode=mode, cache_dir=cache_dir)
    model = models.shakespeare.create_lstm_model()
  elif name == 'STACKOVERFLOW_WORD':
    train, _, test = datasets.stackoverflow.load_data(
        mode=mode, cache_dir=cache_dir)
    tokenizer = datasets.stackoverflow.StackoverflowTokenizer()
    # Same max sequence length as https://arxiv.org/pdf/2003.00295.pdf.
    max_length = 20
    train = train.preprocess_batch(tokenizer.as_preprocess_batch(max_length))
    test = test.preprocess_batch(tokenizer.as_preprocess_batch(max_length))
    model = models.stackoverflow.create_lstm_model(expected_length=13.3)
  else:
    raise ValueError(f'Unsupported task: {name!r}')
  return train, test, model



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Training federated EMNIST via federated averaging.

- The model is a CNN with dropout
- The client optimizer is SGD
- The server optimizer is Adam.

Hyperparameters match those used in https://arxiv.org/abs/2003.00295.
"""

from absl import app

import fedjax
from fedjax.examples import fed_avg

import jax
import jax.numpy as jnp


def main(_):
  # We only use TensorFlow for datasets, so we restrict it to CPU only to avoid
  # issues with certain ops not being available on GPU/TPU.
  fedjax.training.set_tf_cpu_only()

  # Load train and test federated data for EMNIST.
  train_fd, test_fd = fedjax.datasets.emnist.load_data(only_digits=False)

  # Create CNN model with dropout.
  model = fedjax.models.emnist.create_conv_model(only_digits=False)

  # Scalar loss function with model parameters, batch of examples, and seed
  # PRNGKey as input.
  def loss(params, batch, rng):
    # `rng` used with `apply_for_train` to apply dropout during training.
    preds = model.apply_for_train(params, batch, rng)
    # Per example loss of shape [batch_size].
    example_loss = model.train_loss(batch, preds)
    return jnp.mean(example_loss)

  # Gradient function of `loss` w.r.t. to model `params` (jitted for speed).
  grad_fn = jax.jit(jax.grad(loss))

  # Create federated averaging algorithm.
  client_optimizer = fedjax.optimizers.sgd(learning_rate=10**(-1.5))
  server_optimizer = fedjax.optimizers.adam(
      learning_rate=10**(-2.5), b1=0.9, b2=0.999, eps=10**(-4))
  # Hyperparameters for client local traing dataset preparation.
  client_batch_hparams = fedjax.ShuffleRepeatBatchHParams(batch_size=20)
  algorithm = fed_avg.federated_averaging(grad_fn, client_optimizer,
                                          server_optimizer,
                                          client_batch_hparams)

  # Initialize model parameters and algorithm server state.
  init_params = model.init(jax.random.PRNGKey(17))
  server_state = algorithm.init(init_params)

  # Train and eval loop.
  all_client_ids = list(train_fd.client_ids())
  for round_num in range(1500):
    # TODO(b/187167654): Use the ClientSampler interface after it is finalized.
    # Sample 10 clients per round without replacement for training.
    clients = []
    random_state = fedjax.client_samplers.get_pseudo_random_state(
        seed=0, round_num=round_num)
    client_ids = random_state.choice(all_client_ids, size=10, replace=False)
    client_rngs = jax.random.split(jax.random.PRNGKey(round_num), 10)
    for i, (client_id,
            client_dataset) in enumerate(train_fd.get_clients(client_ids)):
      clients.append((client_id, client_dataset, client_rngs[i]))

    # Run one round of training on sampled clients.
    server_state, client_diagnostics = algorithm.apply(server_state, clients)
    print(f'[round {round_num}]')
    # Optionally print client diagnostics if curious about each client's model
    # update's l2 norm.
    # print(f'[round {round_num}] client_diagnostics={client_diagnostics}')

    if round_num % 10 == 0:
      # Periodically evaluate the trained server model parameters.
      # Read and combine clients' train and test datasets for evaluation.
      train_eval_datasets = [cds for _, cds in train_fd.get_clients(client_ids)]
      test_eval_datasets = [cds for _, cds in test_fd.get_clients(client_ids)]
      train_eval_batches = fedjax.padded_batch_client_datasets(
          train_eval_datasets, batch_size=256)
      test_eval_batches = fedjax.padded_batch_client_datasets(
          test_eval_datasets, batch_size=256)

      # Run evaluation metrics defined in `model.eval_metrics`.
      train_metrics = fedjax.evaluate_model(model, server_state.params,
                                            train_eval_batches)
      test_metrics = fedjax.evaluate_model(model, server_state.params,
                                           test_eval_batches)
      print(f'[round {round_num}] train_metrics={train_metrics}')
      print(f'[round {round_num}] test_metrics={test_metrics}')

  # Save final trained model parameters to file.
  fedjax.serialization.save_state(server_state.params, '/tmp/params')

if __name__ == '__main__':
  app.run(main)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Top level experiment script for fed_avg.

Preset hyperparameters can be loaded via -flagfile, e.g.

python3 fed_avg.py -flagfile=fed_avg.EMNIST_CONV.flags -root_dir=/tmp/fed_avg
"""

from typing import Sequence

from absl import app
from absl import flags

import fedjax
import jax


FLAGS = flags.FLAGS
flags.DEFINE_bool('use_parallel', False,
                  'Whether to use jax.pmap fedjax.for_each_client backend.')

task_flags = fedjax.training.structured_flags.TaskFlags()

# Experiment configs.
experiment_config_flags = (
    fedjax.training.structured_flags.FederatedExperimentConfigFlags())
eval_batch_flags = fedjax.training.structured_flags.PaddedBatchHParamsFlags(
    'eval')

# Random seeds.
flags.DEFINE_integer('params_seed', 1,
                     'Seed for initializing model parameters.')
flags.DEFINE_integer('train_sampler_seed', 2,
                     'Seed for the client sampler on the training set.')
flags.DEFINE_integer('test_sampler_seed', 3,
                     'Seed for the client sampler on the test set.')

# Training hyperparameters.
server_optimizer_flags = fedjax.training.structured_flags.OptimizerFlags(
    'server')
client_optimizer_flags = fedjax.training.structured_flags.OptimizerFlags(
    'client')
client_training_batch_flags = (
    fedjax.training.structured_flags.ShuffleRepeatBatchHParamsFlags(
        'client_training'))
flags.DEFINE_integer('num_clients_per_round', 10,
                     'Number of participating clients in each training round.')


def main(argv: Sequence[str]) -> None:
  del argv
  if FLAGS.use_parallel:
    fedjax.set_for_each_client_backend('pmap')

  train_fd, test_fd, model = task_flags.get()
  run_full_periodic_eval = (
      FLAGS.task.startswith('EMNIST_') or FLAGS.task.startswith('SHAKESPEARE_'))

  algorithm = fedjax.algorithms.fed_avg.federated_averaging(
      grad_fn=fedjax.model_grad(model),
      server_optimizer=server_optimizer_flags.get(),
      client_optimizer=client_optimizer_flags.get(),
      client_batch_hparams=client_training_batch_flags.get())

  config = experiment_config_flags.get()

  train_client_sampler = fedjax.client_samplers.UniformGetClientSampler(
      fd=train_fd,
      num_clients=FLAGS.num_clients_per_round,
      seed=FLAGS.train_sampler_seed)
  eval_batch_hparams = eval_batch_flags.get()
  periodic_eval_fn_map = {
      'fed_train_eval':
          fedjax.training.ModelTrainClientsEvaluationFn(model,
                                                        eval_batch_hparams),
      'fed_test_eval':
          fedjax.training.ModelSampleClientsEvaluationFn(
              fedjax.client_samplers.UniformGetClientSampler(
                  fd=test_fd,
                  num_clients=FLAGS.num_clients_per_round,
                  seed=FLAGS.test_sampler_seed), model, eval_batch_hparams)
  }
  final_eval_fn_map = {
      'full_test_eval':
          fedjax.training.ModelFullEvaluationFn(test_fd, model,
                                                eval_batch_hparams)
  }
  if run_full_periodic_eval:
    periodic_eval_fn_map.update(final_eval_fn_map)

  init_state = algorithm.init(model.init(jax.random.PRNGKey(FLAGS.params_seed)))
  fedjax.training.run_federated_experiment(
      algorithm=algorithm,
      init_state=init_state,
      client_sampler=train_client_sampler,
      config=config,
      periodic_eval_fn_map=periodic_eval_fn_map,
      final_eval_fn_map=final_eval_fn_map)


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for working with tree-like container data structures.

In JAX, the term pytree refers to a tree-like structure built out of
container-like Python objects.
For more details, see https://jax.readthedocs.io/en/latest/pytrees.html.
"""

from typing import Iterable, Tuple

from fedjax.core.typing import PyTree

import jax
import jax.numpy as jnp


@jax.jit
def tree_weight(pytree: PyTree, weight: float) -> PyTree:
  """Weights tree leaves by weight."""
  return jax.tree_map(lambda l: l * weight, pytree)


def tree_inverse_weight(pytree: PyTree, weight: float) -> PyTree:
  """Weights tree leaves by ``1 / weight``."""
  inverse_weight = (1. / weight) if weight > 0. else 0.
  return tree_weight(pytree, inverse_weight)


@jax.jit
def tree_zeros_like(pytree: PyTree) -> PyTree:
  """Creates a tree with zeros with same structure as the input."""
  return jax.tree_map(jnp.zeros_like, pytree)


@jax.jit
def tree_add(left: PyTree, right: PyTree) -> PyTree:
  """Adds two trees together."""
  return jax.tree_util.tree_multimap(jnp.add, left, right)


def tree_sum(pytrees: Iterable[PyTree]) -> PyTree:
  """Sums multiple trees together."""
  pytree_sum = None
  for pytree in pytrees:
    if pytree_sum is None:
      pytree_sum = pytree
    else:
      pytree_sum = tree_add(pytree_sum, pytree)
  return pytree_sum


def tree_mean(pytrees_and_weights: Iterable[Tuple[PyTree, float]]) -> PyTree:
  """Returns (weighted) mean of input trees and weights.

  Args:
    pytrees_and_weights: Iterable of tuples of pytrees and associated weights.

  Returns:
    (Weighted) mean of input trees and weights.
  """
  sum_weighted_pytree = None
  sum_weight = 0.
  for pytree, weight in pytrees_and_weights:
    weighted_pytree = tree_weight(pytree, weight)
    if sum_weighted_pytree is None:
      sum_weighted_pytree = weighted_pytree
    else:
      sum_weighted_pytree = tree_add(sum_weighted_pytree, weighted_pytree)
    sum_weight += weight
  return tree_inverse_weight(sum_weighted_pytree, sum_weight)


@jax.jit
def tree_size(pytree: PyTree) -> int:
  """Returns total size of all tree leaves."""
  return sum(l.size for l in jax.tree_util.tree_leaves(pytree))


@jax.jit
def tree_l2_norm(pytree: PyTree) -> float:
  """Returns l2 norm of tree."""
  return jnp.sqrt(
      sum(jnp.vdot(x, x) for x in jax.tree_util.tree_leaves(pytree)))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.federated_data."""

import itertools
import os.path
import zlib

from absl import flags
from absl.testing import absltest
from fedjax.core import federated_data
from fedjax.core import serialization
from fedjax.core import sqlite_federated_data
import numpy as np
import numpy.testing as npt
import sqlite3

FLAGS = flags.FLAGS


class RepeatableIteratorTest(absltest.TestCase):

  def test_two_passes(self):
    # range(5) is repeatable, iter(range(5)) is not.
    it = federated_data.RepeatableIterator(iter(range(5)))
    self.assertListEqual(list(it), [0, 1, 2, 3, 4])
    self.assertListEqual(list(it), [0, 1, 2, 3, 4])

  def test_no_copy_cases(self):
    for values in [[1, 2, 3], (1, 2, 3), '123', b'123', {'1': 2, '3': 4}]:
      it = federated_data.RepeatableIterator(values)
      self.assertIs(values, it._buf)
      self.assertListEqual(list(it), list(values))
      self.assertListEqual(list(it), list(values))


class HelperTest(absltest.TestCase):
  """Tests for helper functions."""

  def test_intersect_slice_ranges(self):
    with self.subTest('start'):
      self.assertEqual(
          federated_data.intersect_slice_ranges(None, None, b'x', None),
          (b'x', None))
      self.assertEqual(
          federated_data.intersect_slice_ranges(b'x', None, None, None),
          (b'x', None))
      self.assertEqual(
          federated_data.intersect_slice_ranges(b'a', None, b'x', None),
          (b'x', None))
      self.assertEqual(
          federated_data.intersect_slice_ranges(b'x', None, b'a', None),
          (b'x', None))

    with self.subTest('stop'):
      self.assertEqual(
          federated_data.intersect_slice_ranges(None, None, None, b'x'),
          (None, b'x'))
      self.assertEqual(
          federated_data.intersect_slice_ranges(None, b'x', None, None),
          (None, b'x'))
      self.assertEqual(
          federated_data.intersect_slice_ranges(None, b'a', None, b'x'),
          (None, b'a'))
      self.assertEqual(
          federated_data.intersect_slice_ranges(None, b'x', None, b'a'),
          (None, b'a'))

    with self.subTest('start and stop'):
      self.assertEqual(
          federated_data.intersect_slice_ranges(None, None, None, None),
          (None, None))
      self.assertEqual(
          federated_data.intersect_slice_ranges(b'a', b'g', b'c', b'h'),
          (b'c', b'g'))


class SubsetFederatedDataTest(absltest.TestCase):

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    path = os.path.join(FLAGS.test_tmpdir, 'test_sqlite_federated_data.sqlite')

    # Create database file. First make sure the database file is empty.
    with open(path, 'w'):
      pass
    connection = sqlite3.connect(path)
    with connection:
      connection.execute("""
      CREATE TABLE federated_data (
        client_id BLOB NOT NULL PRIMARY KEY,
        data BLOB NOT NULL,
        num_examples INTEGER NOT NULL
      );""")
      for i in range(100):
        client_id = f'{i:04x}'.encode('ascii')
        features = {'x': np.arange(i + 1)}
        data = zlib.compress(serialization.msgpack_serialize(features))
        num_examples = i + 1
        connection.execute('INSERT INTO federated_data VALUES (?, ?, ?);',
                           [client_id, data, num_examples])
    connection.close()

    cls.BASE_FEDERATED_DATA = sqlite_federated_data.SQLiteFederatedData.new(
        path)
    cls.CLIENT_IDS = [b'0000', b'000f', b'0012']
    cls.FEDERATED_DATA = federated_data.SubsetFederatedData(
        cls.BASE_FEDERATED_DATA, cls.CLIENT_IDS)

  def test_metadata(self):
    self.assertEqual(self.FEDERATED_DATA.num_clients(), 3)
    self.assertEqual(
        list(self.FEDERATED_DATA.client_ids()), [b'0000', b'000f', b'0012'])
    self.assertEqual(
        list(self.FEDERATED_DATA.client_sizes()), [(b'0000', 1), (b'000f', 16),
                                                   (b'0012', 19)])
    for client_id, size in [(b'0000', 1), (b'000f', 16), (b'0012', 19)]:
      self.assertEqual(self.FEDERATED_DATA.client_size(client_id), size)
    for client_id in [b'xxxx', b'0001', b'1111']:
      with self.assertRaises(KeyError):
        self.FEDERATED_DATA.client_size(client_id)

  def _assert_correct_client_dataset(self,
                                     client_id,
                                     client_dataset,
                                     feature='x',
                                     preprocessor=None):
    self.assertLen(client_id, 4)
    self.assertRegex(client_id, br'00[0-f][0-f]')
    self.assertIn(client_id, self.CLIENT_IDS)
    i = int(client_id, base=16)
    self.assertCountEqual(client_dataset.raw_examples, [feature])
    npt.assert_array_equal(client_dataset.raw_examples[feature],
                           np.arange(i + 1))
    if preprocessor is None:
      self.assertIs(client_dataset.preprocessor,
                    self.BASE_FEDERATED_DATA._preprocess_batch)
    else:
      self.assertIs(client_dataset.preprocessor, preprocessor)

  def test_clients(self):
    client_ids = []
    for client_id, dataset in self.FEDERATED_DATA.clients():
      self.assertIn(client_id, self.CLIENT_IDS)
      self._assert_correct_client_dataset(client_id, dataset)
      client_ids.append(client_id)
    self.assertCountEqual(client_ids, self.CLIENT_IDS)

  def test_shuffled_clients(self):
    client_ids = []
    for client_id, dataset in itertools.islice(
        self.FEDERATED_DATA.shuffled_clients(buffer_size=3, seed=1), 5):
      self.assertIn(client_id, self.CLIENT_IDS)
      self._assert_correct_client_dataset(client_id, dataset)
      client_ids.append(client_id)
    self.assertListEqual(client_ids,
                         [b'0000', b'0012', b'000f', b'000f', b'0012'])

  def test_get_clients(self):
    client_ids = []
    for client_id, dataset in self.FEDERATED_DATA.get_clients(
        [b'0000', b'0012']):
      self.assertIn(client_id, self.CLIENT_IDS)
      self._assert_correct_client_dataset(client_id, dataset)
      client_ids.append(client_id)
    self.assertListEqual(client_ids, [b'0000', b'0012'])

    for client_id in [b'xxxx', b'0001', b'1111']:
      with self.assertRaises(KeyError):
        list(self.FEDERATED_DATA.get_clients([client_id]))

  def test_get_client(self):
    for client_id in self.CLIENT_IDS:
      dataset = self.FEDERATED_DATA.get_client(client_id)
      self._assert_correct_client_dataset(client_id, dataset)

    for client_id in [b'xxxx', b'0001', b'1111']:
      with self.assertRaises(KeyError):
        list(self.FEDERATED_DATA.get_client(client_id))

  def test_slice(self):
    self.assertCountEqual(
        list(self.FEDERATED_DATA.slice().client_ids()),
        [b'0000', b'000f', b'0012'])
    self.assertCountEqual(
        list(self.FEDERATED_DATA.slice(start=b'000f').client_ids()),
        [b'000f', b'0012'])
    self.assertCountEqual(
        list(self.FEDERATED_DATA.slice(stop=b'000f').client_ids()), [b'0000'])
    self.assertCountEqual(
        list(
            self.FEDERATED_DATA.slice(start=b'000f',
                                      stop=b'0012').client_ids()), [b'000f'])

  def test_preprocess_client(self):
    fd = self.FEDERATED_DATA.preprocess_client(lambda _, x: {'y': x['x']})
    self.assertIs(fd._client_ids, self.FEDERATED_DATA._client_ids)
    self.assertIs(fd._base._connection, self.BASE_FEDERATED_DATA._connection)
    self.assertIs(fd._base._preprocess_batch,
                  self.BASE_FEDERATED_DATA._preprocess_batch)
    self._assert_correct_client_dataset(*next(fd.clients()), feature='y')
    self._assert_correct_client_dataset(
        *next(fd.shuffled_clients(buffer_size=10, seed=1)), feature='y')
    self._assert_correct_client_dataset(
        *next(fd.get_clients([b'0000'])), feature='y')
    self._assert_correct_client_dataset(
        b'0012', fd.get_client(b'0012'), feature='y')

  def test_preprocess_batch(self):
    fd = self.FEDERATED_DATA.preprocess_batch(lambda x: {'z': x['x']})
    self.assertIs(fd._client_ids, self.FEDERATED_DATA._client_ids)
    self.assertIs(fd._base._connection, self.BASE_FEDERATED_DATA._connection)
    self.assertIs(fd._base._preprocess_client,
                  self.BASE_FEDERATED_DATA._preprocess_client)
    self.assertCountEqual(
        fd._base._preprocess_batch({'x': np.arange(2)}), ['z'])
    self._assert_correct_client_dataset(
        *next(fd.clients()), preprocessor=fd._base._preprocess_batch)
    self._assert_correct_client_dataset(
        *next(fd.shuffled_clients(buffer_size=10, seed=1)),
        preprocessor=fd._base._preprocess_batch)
    self._assert_correct_client_dataset(
        *next(fd.get_clients([b'0000'])),
        preprocessor=fd._base._preprocess_batch)
    self._assert_correct_client_dataset(
        b'0012',
        fd.get_client(b'0012'),
        preprocessor=fd._base._preprocess_batch)

  def test_bad_client_ids(self):
    with self.assertRaisesRegex(
        ValueError, 'Some client ids are not in the base FederatedData'):
      federated_data.SubsetFederatedData(self.BASE_FEDERATED_DATA, [b'xxxx'])


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.legacy.core.tree_util."""

from fedjax.legacy.core import tree_util
import jax.numpy as jnp
import tensorflow as tf


class TreeUtilTest(tf.test.TestCase):

  def test_tree_broadcast(self):
    pytree = {'x': jnp.array([[0, 0]]), 'y': jnp.array([0])}
    broadcast_pytree = tree_util.tree_broadcast(pytree, axis_size=1)
    self.assertAllEqual(broadcast_pytree['x'], [[[0, 0]]])
    self.assertAllEqual(broadcast_pytree['y'], [[0]])

  def test_tree_stack(self):
    pytrees = [
        {'x': jnp.array([[0, 0]]), 'y': jnp.array([0])},
        {'x': jnp.array([[1, 1]]), 'y': jnp.array([1])},
        {'x': jnp.array([[2, 2]]), 'y': jnp.array([2])},
    ]
    pytree = tree_util.tree_stack(pytrees)
    self.assertAllEqual(pytree['x'], [[[0, 0]], [[1, 1]], [[2, 2]]])
    self.assertAllEqual(pytree['y'], [[0], [1], [2]])

  def test_tree_unstack(self):
    pytree = {
        'x': jnp.array([[[0, 0]], [[1, 1]]]),
        'y': jnp.array([[0], [1]]),
    }
    pytrees = list(tree_util.tree_unstack(pytree, axis_size=2))
    self.assertAllEqual(pytrees[0]['x'], [[0, 0]])
    self.assertAllEqual(pytrees[0]['y'], [0])
    self.assertAllEqual(pytrees[1]['x'], [[1, 1]])
    self.assertAllEqual(pytrees[1]['y'], [1])

  def test_tree_weight(self):
    pytree = {
        'x': jnp.array([[[4, 5]], [[1, 1]]]),
        'y': jnp.array([[3], [1]]),
    }
    weight = 2.0
    weight_pytree = tree_util.tree_weight(pytree, weight)
    self.assertAllEqual(weight_pytree['x'], [[[8.0, 10.0]], [[2.0, 2.0]]])
    self.assertAllEqual(weight_pytree['y'], [[6.0], [2.0]])

  def test_tree_sum(self):
    pytree_1 = {
        'x': jnp.array([[[4, 5]], [[1, 1]]]),
        'y': jnp.array([[3], [1]]),
    }
    pytree_2 = {
        'x': jnp.array([[[2, 3]], [[4, 5]]]),
        'y': jnp.array([[6], [7]]),
    }
    pytree = tree_util.tree_sum(pytree_1, pytree_2)
    self.assertAllEqual(pytree['x'], [[[6, 8]], [[5, 6]]])
    self.assertAllEqual(pytree['y'], [[9], [8]])

  def test_tree_mean(self):
    pytrees = [(0, 1), (2, 3), (4, 5)]
    weights = [6., 7., 8.]
    pytrees_and_weights = zip(pytrees, weights)
    pytree = tree_util.tree_mean(pytrees_and_weights)
    self.assertAllClose(pytree, (2.1904761904761907, 3.1904761904761907))


if __name__ == '__main__':
  tf.test.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logger component."""

import collections
import os.path
from typing import Any, Optional

from absl import logging

import tensorflow as tf


class Logger:
  """Class to encapsulate tf.summary.SummaryWriter logging logic."""

  def __init__(self, root_dir: Optional[str] = None):
    """Initializes summary writers and log directory."""
    self._root_dir = root_dir
    self._summary_writers = collections.OrderedDict()

  def log(self, writer_name: str, metric_name: str, metric_value: Any,
          round_num: int):
    """Records metric using specified summary writer.

    Logs at INFO verbosity. Also, if root_dir is set and metric_value is:
    - a scalar value, convertible to a float32 Tensor, writes scalar summary
    - a vector, convertible to a float32 Tensor, writes histogram summary

    Args:
      writer_name: Name of summary writer.
      metric_name: Name of metric to log.
      metric_value: Value of metric to log.
      round_num: Round number to log.
    """
    logging.info('round %d %s: %s = %s', round_num, writer_name, metric_name,
                 metric_value)

    if self._root_dir is None:
      return

    if writer_name not in self._summary_writers:
      self._summary_writers[writer_name] = tf.summary.create_file_writer(
          os.path.join(self._root_dir, writer_name))

    with self._summary_writers[writer_name].as_default():
      try:
        if getattr(metric_value, 'ndim', 0) > 0:
          tf.summary.histogram(metric_name, metric_value, step=round_num)
        else:
          tf.summary.scalar(metric_name, metric_value, step=round_num)
      except (ValueError, tf.errors.UnimplementedError) as e:
        logging.info('Failed to log summary with exception %s', e)
        pass



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface definitions for federated algorithms."""

import abc
from typing import Generic, List, TypeVar

from fedjax.legacy.core.model import Model
from fedjax.legacy.core.typing import FederatedData

T = TypeVar('T')


class FederatedAlgorithm(Generic[T], metaclass=abc.ABCMeta):
  """Interface for federated algorithms.

  This interface structures federated algorithms at the per round level.
  Algorithm state contains any round specific parameters (e.g. model parameters)
  that will be passed from round to round. This state is initialized by
  `init_state` and passed as input into and returned as output from `run_round`.
  We suggest defining state using `typing.NamedTuple` as this provides
  immutability, type hinting, and works by default with JAX transformations.

  Anything else that is static and doesn't update at each round can be defined
  in the `__init__`.

  The expected usage of FederatedAlgorithm is as follows:

  ```
  algorithm = FederatedAlgorithm()
  state = algorithm.init_state()
  for i in range(num_rounds):
    client_ids = sample_clients(i)
    state = algorithm.run_round(state, client_ids)
  ```
  """

  @abc.abstractproperty
  def federated_data(self) -> FederatedData:
    """Returns federated data used in federated training."""

  @abc.abstractproperty
  def model(self) -> Model:
    """Returns model used in federated training."""

  @abc.abstractmethod
  def init_state(self) -> T:
    """Returns initial state of algorithm."""

  @abc.abstractmethod
  def run_round(self, state: T, client_ids: List[str]) -> T:
    """Runs one round of federated training."""



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transformations and utilities for specifying work for each client.

Many of these utilities work with and require PyTrees as input and output.
This is required when working with the underlying JAX transformations like
`jax.pmap` and `jax.jit`. For more information on what PyTrees are, refer to
https://jax.readthedocs.io/en/latest/pytrees.html.
"""

import abc
import contextlib
import functools
import threading
from typing import Any, Callable, Iterable, Iterator, List, Optional, Sequence, Tuple, Union

from fedjax.core import dataclasses
from fedjax.core.typing import BatchExample
from fedjax.core.typing import PyTree

import jax
import jax.numpy as jnp

# Shared input that is passed to the client init that is shared across all
# clients. For example, this could be the shared global model parameters that
# are then used as the starting point for per client training.
SharedInput = PyTree
# Client specific input passed into client init that is different for each
# client. For example, this could be the starting PRNG seed for training a given
# client. Additionally, for stateful client federated algorithms (usually in the
# cross-silo setting), this is how per client information will be passed along.
ClientInput = PyTree

# Intermittent client state passed as input and output for each client step.
# This will typically contain model parameters and optimizer state that are
# updated at each client step.
ClientStepState = PyTree
# Step results can be used to record any metrics over the course of running the
# for_each_client loop. For example, it can be used to record train metrics for
# each client like gradient norm.
ClientStepResult = PyTree

# Final output from the for_each_client loop. This is usually a subset of the
# client step state. However, more meaningful transformations can be done like
# model update clipping.
ClientOutput = PyTree

ClientInit = Callable[[SharedInput, ClientInput], ClientStepState]
ClientStep = Callable[[ClientStepState, BatchExample], Tuple[ClientStepState,
                                                             ClientStepResult]]
ClientFinal = Callable[[SharedInput, ClientStepState], ClientOutput]

# For the purpose of this module, any hashable type can be used as a client id.
ClientId = Any

ForEachClient = Callable[[
    Iterable[Tuple[ClientId, Iterable[BatchExample], ClientInput]], SharedInput
], Iterable[Tuple[ClientId, ClientOutput, List[ClientStepResult]]]]


class ForEachClientBackend(abc.ABC):

  @abc.abstractmethod
  def __call__(self, client_init: ClientInit, client_step: ClientStep,
               client_final: ClientFinal) -> ForEachClient:
    """Creates a `for_each_client` function."""


class ForEachClientJitBackend(ForEachClientBackend):
  """A straightforward for_each_client backend using jax.jit."""

  def __call__(self, client_init: ClientInit, client_step: ClientStep,
               client_final: ClientFinal) -> ForEachClient:
    """Creates a `for_each_client` function backed by `jax.jit`."""
    jit_client_init = jax.jit(client_init)
    jit_client_step = jax.jit(client_step)
    jit_client_final = jax.jit(client_final)

    def run(shared_input, clients):
      for client_id, client_batches, client_input in clients:
        step_results = []
        state = jit_client_init(shared_input, client_input)
        for batch in client_batches:
          state, step_result = jit_client_step(state, batch)
          step_results.append(step_result)
        output = jit_client_final(shared_input, state)
        yield client_id, output, step_results

    return run


class ForEachClientError(Exception):  # pylint: disable=g-bad-exception-name

  def __init__(self, base: Exception, stage: str, **context):
    super().__init__()
    self.base = base
    self.stage = stage
    self.context = context

  def __str__(self):
    return (f'Stage: {self.stage}. '
            f'Base error is {type(self.base).__name__}: {self.base}.\n'
            'See the `context` field of this exception for additional context.')


class ForEachClientDebugBackend(ForEachClientBackend):
  """for_each_client backend useful during debugging.

  This backend can provide more information for debugging at the cost of being
  slower. With this backend,

  - jax jit compilation is disabled.
  - Exceptions from client_{init,step_final} are wrapped as ForEachClientError
    with the arguments to these functions in the `context` field.
  - Each client is processed sequentially.
  """

  def __call__(self, client_init: ClientInit, client_step: ClientStep,
               client_final: ClientFinal) -> ForEachClient:
    """Creates a `for_each_client` function useful during debugging."""

    def run(shared_input, clients):
      with jax.disable_jit():
        for client_id, client_batches, client_input in clients:
          step_results = []
          try:
            state = client_init(shared_input, client_input)
          except Exception as e:
            raise ForEachClientError(
                e,
                stage='client_init',
                client_id=client_id,
                client_init=client_init,
                shared_input=shared_input,
                client_input=client_input) from e
          for batch in client_batches:
            try:
              state, step_result = client_step(state, batch)
            except Exception as e:
              raise ForEachClientError(
                  e,
                  stage='client_step',
                  client_id=client_id,
                  client_step=client_step,
                  state=state,
                  batch=batch) from e
            step_results.append(step_result)
          try:
            output = client_final(shared_input, state)
          except Exception as e:
            raise ForEachClientError(
                e,
                stage='client_final',
                client_id=client_id,
                client_final=client_final,
                shared_input=shared_input,
                state=state) from e
          yield client_id, output, step_results

    return run


@dataclasses.dataclass
class ClientBlock:
  """Holds data of n clients with padding on both the client and batch levels.

  A ClientBlock holds n uniformly shaped clients, some of which can be "padding
  clients". Further, the final few batches of a client can be "padding batches".

  Attributes:
    client_id: Client ids of each client, or None for a padding client.
    client_mask: Whether each client is a non-padding client.
    num_batches: The number of non-padding batches for each client.
    masked_batches: Uniformly shaped, masked batches for each client. Each list
      of batches for a client has the same length, where the final few batches
      may be padding batches, marked with mask==False.
    client_input: Uniformly shaped client input for each client.
  """
  client_id: List[Optional[ClientId]]
  client_mask: List[bool]
  num_batches: List[int]
  masked_batches: List[List[Tuple[BatchExample, bool]]]
  client_input: List[ClientInput]


def _blockify(clients: Iterable[Tuple[ClientId, Iterable[BatchExample],
                                      ClientInput]],
              block_size: int) -> Iterator[ClientBlock]:
  """Reformats clients into blocks for parallelization.

  Args:
    clients: Iterable of clients, each being a (client_id, client_batches,
      client_input) tuple.
    block_size: Size of output blocks.

  Yields:
    ClientBlocks. The order of clients may be changed in order to minimize the
    amount of padding.
  """
  clients = [(client_id, list(client_batches), client_input)
             for client_id, client_batches, client_input in clients]
  clients.sort(key=lambda x: len(x[1]), reverse=True)
  for i in range(0, len(clients), block_size):
    block = clients[i:i + block_size]
    client_mask = [True for _ in block]
    # Pad to size n.
    if len(block) < block_size:
      _, _, client_input_template = block[0]
      padding_client_input = jax.tree_util.tree_map(jnp.zeros_like,
                                                    client_input_template)
      for j in range(block_size - len(block)):
        # Pad an empty client.
        block.append((None, [], padding_client_input))
        client_mask.append(False)
    # Pad to a fixed number of batches.
    num_batches = [len(client_batches) for _, client_batches, _ in block]
    # Clients are already sorted by decreasing num_batches.
    max_num_batches = num_batches[0]
    masked_batches = []
    if max_num_batches > 0:
      # Use the 0-th batch of the 0-th client in this block as a template.
      batch_template = block[0][1][0]
      padding_batch = jax.tree_util.tree_map(jnp.zeros_like, batch_template)
      for j in range(max_num_batches):
        block_batch = []
        batch_mask = []
        for _, batches, _ in block:
          if j < len(batches):
            block_batch.append(batches[j])
            batch_mask.append(True)
          else:
            block_batch.append(padding_batch)
            batch_mask.append(False)
        masked_batches.append((block_batch, batch_mask))
    yield ClientBlock(
        client_id=[client_id for client_id, _, _ in block],
        client_mask=client_mask,
        num_batches=num_batches,
        masked_batches=masked_batches,
        client_input=[client_input for _, _, client_input in block])


class ForEachClientPmapBackend(ForEachClientBackend):
  """for_each_client backend using jax.pmap for parallelization."""

  def __init__(self, devices: Optional[Sequence[Any]] = None):
    """Initializes a pmap backend.

    Args:
      devices: jax devices to use, defaults to jax.local_devices().
    """
    self._devices = devices

  def __call__(self, client_init: ClientInit, client_step: ClientStep,
               client_final: ClientFinal) -> ForEachClient:
    """Creates a `for_each_client` function using jax.pmap for parallelization.

    Args:
      client_init: ClientInit function.
      client_step: ClientStep function.
      client_final: ClientFinal function.

    Returns:
      Constructed ForEachClient function.
    """
    if self._devices is None:
      devices = jax.local_devices()
    else:
      devices = self._devices
    block_size = len(devices)

    p_client_init = jax.pmap(client_init, in_axes=(None, 0))

    @jax.pmap
    def p_client_step(state, batch, mask):
      next_state, step_result = client_step(state, batch)
      next_state = jax.tree_util.tree_multimap(
          functools.partial(jnp.where, mask), next_state, state)
      step_result = jax.tree_util.tree_map(
          lambda x: jnp.where(mask, x, jnp.zeros_like(x)), step_result)
      return next_state, step_result

    p_client_final = jax.pmap(client_final, in_axes=(None, 0))

    def run(shared_input, clients):
      for block in _blockify(clients, block_size):
        p_state = p_client_init(
            shared_input, jax.device_put_sharded(block.client_input, devices))
        p_step_results = []
        for p_batch, p_mask in block.masked_batches:
          p_state, p_step_result = p_client_step(
              p_state, jax.device_put_sharded(p_batch, devices),
              jax.device_put_sharded(p_mask, devices))
          p_step_results.append(p_step_result)
        p_client_output = p_client_final(shared_input, p_state)
        for i in range(len(block.client_id)):
          if not block.client_mask[i]:
            continue
          client_output, step_results = jax.tree_util.tree_map(
              lambda x: x[i],  # pylint: disable=cell-var-from-loop
              (p_client_output, p_step_results))
          yield (block.client_id[i], client_output,
                 step_results[:block.num_batches[i]])

    return run


class BackendChoice(threading.local):
  """Thread local configuration of the for_each_client backend choice."""

  DEFAULT_BACKEND = ForEachClientJitBackend()

  def __init__(self):
    super().__init__()
    self.backend = None

  def get(self):
    if self.backend is None:
      self.backend = self.DEFAULT_BACKEND
    return self.backend


_BACKEND_CHOICE = BackendChoice()


def get_for_each_client_backend() -> ForEachClientBackend:
  return _BACKEND_CHOICE.get()


def set_for_each_client_backend(backend: Union[ForEachClientBackend, str,
                                               None]):
  """Sets the for_each_client backend for the current thread.

  Args:
    backend: One of the following,

      - None: uses the default backend for the current environment.
      - 'debug': uses the debugging backend.
      - 'jit': uses the JIT backend.
      - 'pmap': uses the pmap-based backend.
      - A concrete ForEachClientBackend object.
  """
  if backend is None or isinstance(backend, ForEachClientBackend):
    _BACKEND_CHOICE.backend = backend
  elif backend == 'debug':
    _BACKEND_CHOICE.backend = ForEachClientDebugBackend()
  elif backend == 'jit':
    _BACKEND_CHOICE.backend = ForEachClientJitBackend()
  elif backend == 'pmap':
    _BACKEND_CHOICE.backend = ForEachClientPmapBackend()
  else:
    raise ValueError(f'Unsupported backend {backend!r}')


@contextlib.contextmanager
def for_each_client_backend(backend: Union[ForEachClientBackend, str, None]):
  """A context manager for switching to a given ForEachClientBackend in the current thread.

  Example::

    with for_each_client_backend('pmap'):
      # We will be using the pmap based for_each_client backend within this block.
      pass
    # We will be using the default for_each_client backend from now on.

  Args:
    backend: See :func:`set_for_each_client_backend`.

  Yields:
    Nothing.
  """
  try:
    old = _BACKEND_CHOICE.backend
    set_for_each_client_backend(backend)
    yield
  finally:
    set_for_each_client_backend(old)


# We leave the return type unannotated because there's no easy way to properly
# annotate it when it depends on the input value of with_step_result.
def for_each_client(client_init: ClientInit,
                    client_step: Union[ClientStep,
                                       Callable[[ClientStepState, BatchExample],
                                                ClientStepState]],
                    client_final: ClientFinal = lambda _, s: s,
                    with_step_result: bool = False):
  """Creates a function which maps over clients.

  For example, for_each_client could be used to define how to run client
  updates for each client in a federated training round. Another common use case
  of for_each_client is to run evaluation per client for a given set of model
  parameters.

  The underlying backend for for_each_client is customizable.
  For example, if multiple devies are available (e.g. TPU), a :func:`jax.pmap`
  based backend can be used to parallelize across devices.
  It's also possible to manually specify which backend to use (for debugging).

  The expected usage of for_each_client is as follows::

    # Map over clients and count how many points are greater than `limit` for
    # each client. Each client also has a different `start` that is specified
    # via client input.

    def client_init(shared_input, client_input):
      client_step_state = {
          'limit': shared_input['limit'],
          'count': client_input['start']
      }
      return client_step_state

    def client_step(client_step_state, batch):
      num = jnp.sum(batch['x'] > client_step_state['limit'])
      client_step_state = {
          'limit': client_step_state['limit'],
          'count': client_step_state['count'] + num
      }
      return client_step_state

    def client_final(shared_input, client_step_state):
      del shared_input  # Unused.
      return client_step_state['count']

    # Three clients with different data and starting counts.
    # clients = [(client_id, client_batches, client_input)]
    clients = [
        (b'cid0',
        [{'x': jnp.array([1, 2, 3, 4])}, {'x': jnp.array([1, 2, 3])}],
        {'start': jnp.array(2)}),
        (b'cid1',
        [{'x': jnp.array([1, 2])}, {'x': jnp.array([1, 2, 3, 4, 5])}],
        {'start': jnp.array(0)}),
        (b'cid2',
        [{'x': jnp.array([1])}],
        {'start': jnp.array(1)}),
    ]
    shared_input = {'limit': jnp.array(2)}

    func = fedjax.for_each_client.for_each_client(
        client_init, client_step, client_final)
    print(list(func(shared_input, clients)))
    # [(b'cid0', 5), (b'cid1', 3), (b'cid2', 1)]

  Here's the same example with per step results. ::

    # We'll also keep track of the `num` per step in our step results.

    def client_step_with_result(client_step_state, batch):
      num = jnp.sum(batch['x'] > client_step_state['limit'])
      client_step_state = {
          'limit': client_step_state['limit'],
          'count': client_step_state['count'] + num
      }
      client_step_result = {'num': num}
      return client_step_state, client_step_result

    func = fedjax.for_each_client.for_each_client(
        client_init, client_step_with_result, client_final, with_step_result=True)
    print(list(func(shared_input, clients)))
    # [
    #   (b'cid0', 5, [{'num': 2}, {'num': 1}]),
    #   (b'cid1', 3, [{'num': 0}, {'num': 3}]),
    #   (b'cid2', 1, [{'num': 0}]),
    # ]

  Args:
    client_init: Function that initializes the internal intermittent client step
      state from the share input and per client input. The shared input contains
      information like the global model parameters that are shared across all
      clients. The per client input is per client information. The initialized
      internal client step state is fed as intermittent input and output from
      `client_step` and `client_final`. This client step state usually contains
      the model parameters and optimizer state for each client that are updated
      at each `client_step`. This will be run once for each client.
    client_step: Function that takes the client step state and a batch of
      examples as input and outputs a (possibly updated) client step state.
      Optionally, per step results can also be returned as the second element if
      `with_step_result` is True. Per step results are usually diagnostics like
      gradient norm. This will be run for each batch for each client.
    client_final: Function that applies the final transformation on the internal
      client step state to the desired final client output. More meaningful
      transformations can be done here, like model update clipping. Defaults to
      just returning the client step state. This will be run once for each
      client.
    with_step_result: Indicates whether client_step returns a pair where the
      first element is considered the client output and the second element is
      the client step result.

  Returns:
    A for each client function that takes shared_input and the per client inputs
    as tuple (client_id, batched_client_data, client_rng) to map over and
    returns the outputs per client as specified in `client_final` along with
    optional per client per step results.
  """
  for_each_client_backend_ = get_for_each_client_backend()
  if with_step_result:
    return for_each_client_backend_(client_init, client_step, client_final)

  def client_step_with_result(client_step_state, batch):
    return client_step(client_step_state, batch), ()

  func = for_each_client_backend_(client_init, client_step_with_result,
                                  client_final)

  def run(shared_input, clients):
    for client_id, client_output, _ in func(shared_input, clients):
      yield client_id, client_output

  return run



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.metrics."""

from absl.testing import absltest
from absl.testing import parameterized

from fedjax.core import metrics

import jax.numpy as jnp
import numpy.testing as npt


class MeanStatTest(absltest.TestCase):

  def test_str(self):
    stat = metrics.MeanStat.new(2, 4)
    self.assertEqual(
        'MeanStat(accum=DeviceArray(2, dtype=int32), weight=DeviceArray(4, dtype=int32)) => 0.5',
        str(stat))

  def test_new(self):
    stat = metrics.MeanStat.new(jnp.array([2, 3, 1]), jnp.array([1, 0, 1]))
    npt.assert_array_equal(stat.accum, [2, 0, 1])
    npt.assert_array_equal(stat.weight, [1, 0, 1])

  def test_result(self):
    stat = metrics.MeanStat.new(2, 5)
    self.assertEqual(stat.result(), 0.4)

  def test_merge(self):
    stat_0 = metrics.MeanStat.new(1, 2)
    stat_1 = metrics.MeanStat.new(2, 3)
    merged_stat = stat_0.merge(stat_1)
    self.assertEqual(merged_stat.accum, 3)
    self.assertEqual(merged_stat.weight, 5)

  def test_reduce(self):
    stat = metrics.MeanStat.new(jnp.array([1, 2, 4]), jnp.array([1, 1, 0]))
    reduced_stat = stat.reduce()
    self.assertEqual(reduced_stat.accum, 3)
    self.assertEqual(reduced_stat.weight, 2)


class SumStatTest(absltest.TestCase):

  def test_str(self):
    stat = metrics.SumStat.new(2)
    self.assertEqual('SumStat(accum=DeviceArray(2, dtype=int32)) => 2',
                     str(stat))

  def test_result(self):
    stat = metrics.SumStat.new(2)
    self.assertEqual(stat.result(), 2)

  def test_merge(self):
    stat_0 = metrics.SumStat.new(1)
    stat_1 = metrics.SumStat.new(2)
    merged_stat = stat_0.merge(stat_1)
    self.assertEqual(merged_stat.accum, 3)

  def test_reduce(self):
    stat = metrics.SumStat.new(jnp.array([1, 2, 1]))
    reduced_stat = stat.reduce()
    self.assertEqual(reduced_stat.accum, 4)


class MetricsTest(parameterized.TestCase):

  def test_cross_entropy_loss(self):
    example = {'y': jnp.array(1)}
    prediction = jnp.array([1.2, 0.4])
    metric = metrics.CrossEntropyLoss()
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
      self.assertEqual(zero.weight, 0)
    with self.subTest('evaluate_example'):
      loss = metric.evaluate_example(example, prediction)
      self.assertAlmostEqual(loss.accum, 1.1711007)
      self.assertAlmostEqual(loss.weight, 1)

  @parameterized.named_parameters(
      {
          'testcase_name': 'correct',
          'target': 2,
          'prediction': [0, 0, 1],
          'expected_result': 1.,
      }, {
          'testcase_name': 'incorrect',
          'target': 1,
          'prediction': [1, 0, 0],
          'expected_result': 0.,
      })
  def test_accuracy(self, target, prediction, expected_result):
    example = {'y': jnp.array(target)}
    prediction = jnp.array(prediction)
    metric = metrics.Accuracy()
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
      self.assertEqual(zero.weight, 0)
    with self.subTest('evaluate_example'):
      accuracy = metric.evaluate_example(example, prediction)
      self.assertEqual(accuracy.result(), expected_result)

  @parameterized.named_parameters(
      {
          'testcase_name': 'correct, 0 < k < classes',
          'kval': 3,
          'target': 2,
          'prediction': [0.7, 0., 0.2, -0.02, 0.6],
          'expected_result': 1.,
      }, {
          'testcase_name': 'incorrect, 0 < k < classes',
          'kval': 2,
          'target': 0,
          'prediction': [0., -0.2, 0.5, 0.55, 0.],
          'expected_result': 0.,
      }, {
          'testcase_name': 'incorrect, k = 0',
          'kval': 0,
          'target': 2,
          'prediction': [0., 0., 1., 0.],
          'expected_result': 0.,
      }, {
          'testcase_name': 'incorrect, k < 0',
          'kval': -5,
          'target': 2,
          'prediction': [0., 0., 1., 0.],
          'expected_result': 0.,
      }, {
          'testcase_name': 'correct, k > classes',
          'kval': 5,
          'target': 2,
          'prediction': [0., 0., 1., 0.],
          'expected_result': 1.,
      }, {
          'testcase_name': 'correct, same prediction for multiple classes',
          'kval': 3,
          'target': 0,
          'prediction': [0., .3, 1., 0.],
          'expected_result': 1.,
      }, {
          'testcase_name': 'incorrect, same prediction for multiple classes',
          'kval': 3,
          'target': 3,
          'prediction': [0., 0., 1., 0.],
          'expected_result': 0.,
      })
  def test_top_k_accuracy(self, kval, target, prediction, expected_result):
    example = {'y': jnp.array(target)}
    prediction = jnp.array(prediction)
    metric = metrics.TopKAccuracy(k=kval)
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
      self.assertEqual(zero.weight, 0)
    with self.subTest('evaluate_example'):
      top_k_accuracy = metric.evaluate_example(example, prediction)
      self.assertEqual(top_k_accuracy.accum, expected_result)
      self.assertEqual(top_k_accuracy.weight, 1.)

  def test_sequence_token_cross_entropy_loss(self):
    example = {'y': jnp.array([1, 0, 1])}
    prediction = jnp.array([[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]])
    metric = metrics.SequenceTokenCrossEntropyLoss()
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
      self.assertEqual(zero.weight, 0)
    with self.subTest('evaluate_example'):
      loss = metric.evaluate_example(example, prediction)
      self.assertAlmostEqual(loss.accum, 1.2246635)
      self.assertAlmostEqual(loss.weight, 2)
    with self.subTest('per_position evaluate_example'):
      per_position_metric = metrics.SequenceTokenCrossEntropyLoss(
          per_position=True)
      per_position_loss = per_position_metric.evaluate_example(
          example, prediction)
      npt.assert_array_almost_equal(per_position_loss.accum,
                                    [1.1711007, 0., 0.05356275])
      npt.assert_array_equal(per_position_loss.weight, [1., 0., 1.])

  def test_sequence_cross_entropy_loss(self):
    example = {'y': jnp.array([1, 0, 1])}
    prediction = jnp.array([[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]])
    metric = metrics.SequenceCrossEntropyLoss()
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
      self.assertEqual(zero.weight, 0)
    with self.subTest('evaluate_example'):
      loss = metric.evaluate_example(example, prediction)
      self.assertAlmostEqual(loss.accum, 1.2246635)
      self.assertAlmostEqual(loss.weight, 1)

  def test_sequence_token_accuracy(self):
    example = {'y': jnp.array([1, 2, 2, 1, 3, 0])}
    # prediction = [1, 0, 2, 1, 3, 0].
    prediction = jnp.array([[0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 1, 0],
                            [0, 1, 0, 0], [0, 0, 0, 1], [1, 0, 0, 0]])
    logits_mask = (0., 0., 0., jnp.NINF)
    metric = metrics.SequenceTokenAccuracy(logits_mask=logits_mask)
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
      self.assertEqual(zero.weight, 0)
    with self.subTest('evaluate_example'):
      accuracy = metric.evaluate_example(example, prediction)
      self.assertEqual(accuracy.accum, 3)
      self.assertEqual(accuracy.weight, 5)
    with self.subTest('per_position evaluate_example'):
      per_position_metric = metrics.SequenceTokenAccuracy(
          logits_mask=logits_mask, per_position=True)
      per_position_accuracy = per_position_metric.evaluate_example(
          example, prediction)
      npt.assert_array_almost_equal(per_position_accuracy.accum,
                                    [1., 0., 1., 1., 0., 0.])
      npt.assert_array_equal(per_position_accuracy.weight,
                             [1., 1., 1., 1., 1., 0.])

  def test_sequence_token_top_k_accuracy(self):
    example = {'y': jnp.array([1, 2, 2, 1, 3, 0])}
    prediction = jnp.array([[0, 1, 0.5, 0], [1, 0.5, 0, 0], [0.8, 0, 0.7, 0.2],
                            [0.5, 1, 0, 1], [0, 0.5, 0, 1], [0.5, 0, 0.9, 0.5]])
    logits_mask = (0., 0., 0., jnp.NINF)
    metric = metrics.SequenceTokenTopKAccuracy(k=2, logits_mask=logits_mask)
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
      self.assertEqual(zero.weight, 0)
    with self.subTest('evaluate_example'):
      accuracy = metric.evaluate_example(example, prediction)
      self.assertEqual(accuracy.accum, 3)
      self.assertEqual(accuracy.weight, 5)
    with self.subTest('per_position evaluate_example'):
      per_position_metric = metrics.SequenceTokenTopKAccuracy(
          k=2, logits_mask=logits_mask, per_position=True)
      per_position_accuracy = per_position_metric.evaluate_example(
          example, prediction)
      npt.assert_array_almost_equal(per_position_accuracy.accum,
                                    [1., 0., 1., 1., 0., 0.])
      npt.assert_array_equal(per_position_accuracy.weight,
                             [1., 1., 1., 1., 1., 0.])

  def test_sequence_token_count(self):
    example = {'y': jnp.array([1, 2, 2, 3, 4, 0, 0])}
    prediction = jnp.array([])  # Unused.
    metric = metrics.SequenceTokenCount(masked_target_values=(0, 2))
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
    with self.subTest('evaluate_example'):
      count = metric.evaluate_example(example, prediction)
      self.assertEqual(count.accum, 3)

  def test_sequence_count(self):
    example = {'y': jnp.array([1, 2, 2, 3, 4, 0, 0])}
    empty_example = {'y': jnp.array([0, 0, 0, 0, 0, 0, 0])}
    prediction = jnp.array([])  # Unused.
    metric = metrics.SequenceCount(masked_target_values=(0, 2))
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
    with self.subTest('evaluate_example'):
      self.assertEqual(metric.evaluate_example(example, prediction).accum, 1)
      self.assertEqual(
          metric.evaluate_example(empty_example, prediction).accum, 0)

  @parameterized.named_parameters(
      {
          'testcase_name': 'untruncated',
          'target': [1, 2, 2, 3, 4, 0, 0],
          'expected_result': 0.,
      }, {
          'testcase_name': 'truncated',
          'target': [1, 2, 2, 3, 3, 3, 3],
          'expected_result': 1.,
      })
  def test_sequence_truncation_rate(self, target, expected_result):
    example = {'y': jnp.array(target)}
    prediction = jnp.array([])  # Unused.
    metric = metrics.SequenceTruncationRate(eos_target_value=4)
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
      self.assertEqual(zero.weight, 0)
    with self.subTest('evaluate_example'):
      truncation_rate = metric.evaluate_example(example, prediction)
      self.assertEqual(truncation_rate.result(), expected_result)

  def test_sequence_token_oov_rate(self):
    example = {'y': jnp.array([1, 2, 2, 3, 4, 0, 0])}
    prediction = jnp.array([])  # Unused.
    metric = metrics.SequenceTokenOOVRate(oov_target_values=(2,))
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
      self.assertEqual(zero.weight, 0)
    with self.subTest('evaluate_example'):
      oov_rate = metric.evaluate_example(example, prediction)
      self.assertEqual(oov_rate.accum, 2)
      self.assertEqual(oov_rate.weight, 5)
    with self.subTest('per_position evaluate_example'):
      per_position_metric = metrics.SequenceTokenOOVRate(
          oov_target_values=(2,), per_position=True)
      per_position_oov_rate = per_position_metric.evaluate_example(
          example, prediction)
      npt.assert_array_almost_equal(per_position_oov_rate.accum,
                                    [0., 1., 1., 0., 0., 0., 0.])
      npt.assert_array_equal(per_position_oov_rate.weight,
                             [1., 1., 1., 1., 1., 0., 0.])

  def test_sequence_length(self):
    example = {'y': jnp.array([1, 2, 3, 4, 0, 0])}
    prediction = jnp.array([])  # Unused.
    metric = metrics.SequenceLength()
    with self.subTest('zero'):
      zero = metric.zero()
      self.assertEqual(zero.accum, 0)
      self.assertEqual(zero.weight, 0)
    with self.subTest('evaluate_example'):
      sequence_length = metric.evaluate_example(example, prediction)
      self.assertEqual(sequence_length.accum, 4)
      self.assertEqual(sequence_length.weight, 1)

  def test_per_domain(self):
    metric = metrics.PerDomainMetric(metrics.Accuracy(), num_domains=4)

    stat = metric.zero()
    self.assertIsInstance(stat, metrics.MeanStat)
    npt.assert_array_equal(stat.accum, [0., 0., 0., 0.])
    npt.assert_array_equal(stat.weight, [0., 0., 0., 0.])

    stat = metric.evaluate_example({
        'y': jnp.array(1),
        'domain_id': 0
    }, jnp.array([0., 1.]))
    self.assertIsInstance(stat, metrics.MeanStat)
    npt.assert_array_equal(stat.accum, [1., 0., 0., 0.])
    npt.assert_array_equal(stat.weight, [1., 0., 0., 0.])

    stat = metric.evaluate_example({
        'y': jnp.array(0),
        'domain_id': 2
    }, jnp.array([0., 1.]))
    self.assertIsInstance(stat, metrics.MeanStat)
    npt.assert_array_equal(stat.accum, [0., 0., 0., 0.])
    npt.assert_array_equal(stat.weight, [0., 0., 1., 0.])

  @parameterized.named_parameters(
      {
          'testcase_name': 'correct',
          'target': 1,
          'prediction': [0., 1., 0.],
          'expected_result': [[0., 0., 0.], [0., 1., 0.], [0., 0., 0.]],
      }, {
          'testcase_name': 'incorrect',
          'target': 1,
          'prediction': [1., 0., 0.],
          'expected_result': [[0., 0., 0.], [1., 0., 0.], [0., 0., 0.]],
      })
  def test_confusion_matrix(self, target, prediction, expected_result):
    example = {'y': jnp.array(target)}
    prediction = jnp.array(prediction)
    num_classes = len(prediction)
    metric = metrics.ConfusionMatrix(num_classes=num_classes)
    with self.subTest('zero'):
      zero = metric.zero()
      npt.assert_array_equal(zero.accum, jnp.zeros((num_classes,
                                                    num_classes)))
    with self.subTest('evaluate_example'):
      confusion_matrix = metric.evaluate_example(example, prediction)
      npt.assert_array_equal(confusion_matrix.result(), expected_result)

  def test_confusion_matrix_invalid_input(self):
    example = {'y': jnp.array(1)}
    prediction = jnp.array([0., 1.])
    metric = metrics.ConfusionMatrix(num_classes=3)
    with self.assertRaisesRegex(ValueError, 'Make sure num_classes is equal'):
      metric.evaluate_example(example, prediction)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Toy regression models."""

from fedjax.core import models

import haiku as hk
import jax.numpy as jnp
import numpy as np


def create_regression_model() -> models.Model:
  """Creates toy regression model.

  Matches the model used in:

  Communication-Efficient Agnostic Federated Averaging
    Jae Ro, Mingqing Chen, Rajiv Mathews, Mehryar Mohri, Ananda Theertha Suresh
    https://arxiv.org/abs/2104.02748

  Returns:
    Model
  """

  def forward_pass(batch):
    network = hk.Sequential([hk.Linear(1, with_bias=False)])
    return jnp.mean(network(batch['x']))

  def train_loss(batch, preds):
    return jnp.square(jnp.mean(batch['y']) - preds)

  transformed_forward_pass = hk.transform(forward_pass)
  sample_batch = {'x': np.zeros((1, 1)), 'y': np.zeros((1,))}
  return models.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=sample_batch,
      train_loss=train_loss)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.legacy.core.model."""

import collections

from fedjax.legacy.core import metrics
from fedjax.legacy.core import model
import haiku as hk
import jax
from jax.experimental import stax
import numpy as np
import tensorflow as tf


def _loss(batch, preds):
  return metrics.cross_entropy_loss_fn(targets=batch['y'], preds=preds)


def _accuracy(batch, preds):
  return metrics.accuracy_fn(targets=batch['y'], preds=preds)


def _create_haiku_model(num_classes,
                        sample_batch,
                        non_trainable_module_names=()):
  """Creates toy haiku model."""

  def forward_pass(batch):
    network = hk.Sequential(
        [hk.Linear(2 * num_classes),
         hk.Linear(num_classes)])
    return network(batch['x'])

  transformed_forward_pass = hk.transform(forward_pass)
  metrics_fn_map = collections.OrderedDict(accuracy=_accuracy)
  return model.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=sample_batch,
      loss_fn=_loss,
      metrics_fn_map=metrics_fn_map,
      non_trainable_module_names=non_trainable_module_names)


def _create_stax_model(num_classes, sample_shape):
  """Creates toy stax model."""
  stax_init_fn, stax_apply_fn = stax.serial(stax.Flatten,
                                            stax.Dense(2 * num_classes),
                                            stax.Dense(num_classes))
  metrics_fn_map = collections.OrderedDict(accuracy=_accuracy)
  return model.create_model_from_stax(
      stax_init_fn=stax_init_fn,
      stax_apply_fn=stax_apply_fn,
      sample_shape=sample_shape,
      loss_fn=_loss,
      metrics_fn_map=metrics_fn_map)


class StaxModelTest(tf.test.TestCase):

  def create_model(self):
    return _create_stax_model(self._num_classes, (-1, self._feature_dim))

  def setUp(self):
    super().setUp()
    self._batch_size = 5
    self._feature_dim = 10
    self._num_classes = 2
    self._rng_seq = hk.PRNGSequence(0)
    self._batch = collections.OrderedDict(
        x=np.ones((self._batch_size, self._feature_dim)),
        y=np.ones((self._batch_size,)))
    self._model = self.create_model()

  def test_init_params(self):
    params = self._model.init_params(rng=next(self._rng_seq))
    self.assertEqual(hk.data_structures.tree_size(params), 54)

  def test_backward_pass(self):
    params = self._model.init_params(rng=next(self._rng_seq))
    output = self._model.backward_pass(
        params=params, batch=self._batch, rng=next(self._rng_seq))
    tf.nest.map_structure(lambda a, b: self.assertTupleEqual(a.shape, b.shape),
                          params, output.grads)
    self.assertGreaterEqual(output.loss, 0.)
    self.assertEqual(output.num_examples, self._batch_size)

  def test_evaluate(self):
    params = self._model.init_params(rng=next(self._rng_seq))
    eval_metrics = self._model.evaluate(params=params, batch=self._batch)

    self.assertContainsSubset(
        ['num_examples', 'regularizer', 'loss', 'accuracy'],
        list(eval_metrics.keys()))

  def test_train(self):
    params = self._model.init_params(rng=next(self._rng_seq))
    prev_loss = self._model.evaluate(
        params=params, batch=self._batch)['loss'].result()
    for _ in range(5):
      output = self._model.backward_pass(
          params=params, batch=self._batch, rng=next(self._rng_seq))
      params = jax.tree_multimap(lambda p, g: p - 0.01 * g, params,
                                 output.grads)
      loss = self._model.evaluate(
          params=params, batch=self._batch)['loss'].result()
      self.assertLess(loss, prev_loss)
      prev_loss = loss


class HaikuModelTest(StaxModelTest):

  def create_model(self):
    return _create_haiku_model(self._num_classes, self._batch)

  def test_train_with_some_frozen(self):
    non_trainable_module_names = ['linear_1']
    model_ = _create_haiku_model(self._num_classes, self._batch,
                                 non_trainable_module_names)
    params = model_.init_params(rng=next(self._rng_seq))
    predicate = lambda module_name, *_: module_name in non_trainable_module_names
    init_non_trainable_params, init_trainable_params = (
        hk.data_structures.partition(predicate, params))

    for _ in range(5):
      output = model_.backward_pass(
          params=params, batch=self._batch, rng=next(self._rng_seq))
      params = jax.tree_multimap(lambda p, g: p - 0.01 * g, params,
                                 output.grads)

    non_trainable_params, trainable_params = (
        hk.data_structures.partition(predicate, params))
    jax.tree_multimap(self.assertAllEqual, non_trainable_params,
                      init_non_trainable_params)
    jax.tree_multimap(self.assertNotAllEqual, trainable_params,
                      init_trainable_params)


if __name__ == '__main__':
  tf.test.main()
